import os
import sys

from strictdoc.backend.dsl.reader import SDReader
from strictdoc.backend.dsl.writer import SDWriter


class PassthroughAction:

    def passthrough(self, path_to_doc, output_file):
        if not os.path.isfile(path_to_doc):
            sys.stdout.flush()
            err = "Could not open doc file '{}': No such file or directory".format(path_to_doc)
            print(err)
            exit(1)

        document = SDReader().read_from_file(path_to_doc)

        writer = SDWriter()
        output = writer.write(document)
        with open(output_file, 'w') as file:
            file.write(output)



from docutils.core import publish_parts


class RstToHtmlFragmentWriter:

    @staticmethod
    def write(rst_fragment):
        # How do I convert a docutils document tree into an HTML string?
        # https://stackoverflow.com/a/32168938/598057
        html = publish_parts(rst_fragment, writer_name='html')['html_body']
        return html



from strictdoc.backend.dsl.models.requirement import Requirement
from strictdoc.core.document_iterator import DocumentCachingIterator
from strictdoc.core.document_tree import DocumentTree
from strictdoc.helpers.sorting import alphanumeric_sort


class TraceabilityIndex:
    @staticmethod
    def create(document_tree: DocumentTree):
        requirements_map = {}
        requirements_children_map = {}
        tags_map = {}
        document_iterators = {}

        # It seems to be impossible to accomplish everything in just one for
        # loop. One particular problem that requires two passes: it is not
        # possible to know after one iteration if any of the requirements
        # reference parents that do not exist.
        #
        # Step #1:
        # - Collect a dictionary of all requirements in the document tree:
        # {req_id: req}
        # - Each requirement's 'parents_uids' is populated with the forward
        # declarations of its parents uids.
        # - A separate map is created: {req_id: [req_children]}
        # At this point some information is in place but it was not known if
        # some of the UIDs could not be resolved which is the task of the second
        # step.
        #
        # Step #2:
        # - Check if each requirement's has valid parent links.
        # - Resolve parent forward declarations
        # - Re-assign children declarations
        # - Calculate depth of both parent and child links.
        for document in document_tree.document_list:
            document_iterator = DocumentCachingIterator(document)
            document_iterators[document] = document_iterator

            if document.name not in tags_map:
                tags_map[document.name] = {}

            for section_or_requirement in document_iterator.all_content():
                if not section_or_requirement.is_requirement:
                    continue

                requirement: Requirement = section_or_requirement
                if not requirement.uid:
                    continue

                document_tags = tags_map[document.name]
                for tag in requirement.tags:
                    if tag not in document_tags:
                        document_tags[tag] = 0
                    document_tags[tag] += 1

                if requirement.uid in requirements_map:
                    print('error: DocumentIndex: requirement already exists: {}'.format(
                        requirement.uid
                    ))
                    exit(1)

                if requirement.uid not in requirements_children_map:
                    requirements_children_map[requirement.uid] = []

                requirements_map[requirement.uid] = {
                    'document': document,
                    'requirement': requirement,
                    'parents': [],
                    'parents_uids': [],
                    'children': []
                }

                for ref in requirement.references:
                    if ref.ref_type != "Parent":
                        continue
                    requirements_map[requirement.uid]['parents_uids'].append(ref.path)

                    if ref.path not in requirements_children_map:
                        requirements_children_map[ref.path] = []
                    requirements_children_map[ref.path].append(requirement)

        # Now iterate over the requirements again to build an in-depth map of
        # parents and children.
        requirements_child_depth_map = {}
        requirements_parent_depth_map = {}
        documents_ref_depth_map = {}

        for document in document_tree.document_list:
            document_iterator = document_iterators[document]
            max_parent_depth, max_child_depth = 0, 0

            for section_or_requirement in document_iterator.all_content():
                if not section_or_requirement.is_requirement:
                    continue

                requirement: Requirement = section_or_requirement
                if not requirement.uid:
                    continue

                # Now it is possible to resolve parents first checking if they
                # indeed exist.
                requirement_parent_ids = requirements_map[requirement.uid]['parents_uids']
                for requirement_parent_id in requirement_parent_ids:
                    if requirement_parent_id not in requirements_map:
                        # TODO: Strict variant of the behavior will be to stop
                        # and raise an error message.
                        print("warning: [DocumentIndex.create] Requirement {} references parent requirement which doesn't exist: {}".format(
                            requirement.uid, requirement_parent_id
                        ))
                        requirements_children_map.pop(requirement_parent_id, None)
                        continue
                    parent_requirement = requirements_map[requirement_parent_id]['requirement']
                    requirements_map[requirement.uid]['parents'].append(parent_requirement)

                if requirement.uid not in requirements_child_depth_map:
                    child_depth = 0

                    requirements_map[requirement.uid]['children'] = requirements_children_map[requirement.uid]

                    queue = requirements_children_map[requirement.uid]
                    while True:
                        if len(queue) == 0:
                            break

                        child_depth += 1
                        deeper_queue = []
                        for child in queue:
                            deeper_queue.extend(requirements_children_map[child.uid])
                        queue = deeper_queue

                    requirements_child_depth_map[requirement.uid] = child_depth
                    if max_child_depth < child_depth:
                        max_child_depth = child_depth

                # Calculate parent depth
                if requirement.uid not in requirements_parent_depth_map:
                    parent_depth = 0

                    queue = requirement_parent_ids
                    while True:
                        if len(queue) == 0:
                            break

                        parent_depth += 1
                        deeper_queue = []
                        for parent_uid in queue:
                            if parent_uid not in requirements_map:
                                continue
                            deeper_queue.extend(requirements_map[parent_uid]['parents_uids'])
                        queue = deeper_queue

                    requirements_parent_depth_map[requirement.uid] = parent_depth
                    if max_parent_depth < parent_depth:
                        max_parent_depth = parent_depth

            documents_ref_depth_map[document] = max(max_parent_depth, max_child_depth)

        traceability_index = TraceabilityIndex(
            document_iterators,
            requirements_map,
            tags_map,
            documents_ref_depth_map
        )
        return traceability_index

    def __init__(self,
                 document_iterators,
                 requirements_parents,
                 tags_map,
                 documents_ref_depth_map):
        self._document_iterators = document_iterators
        self._requirements_parents = requirements_parents
        self._tags_map = tags_map
        self._documents_ref_depth_map = documents_ref_depth_map

    @property
    def document_iterators(self):
        return self._document_iterators

    @property
    def requirements_parents(self):
        return self._requirements_parents

    @property
    def tags_map(self):
        return self._tags_map

    @property
    def documents_ref_depth_map(self):
        return self._documents_ref_depth_map

    def get_document_iterator(self, document):
        return self.document_iterators[document]

    def get_parent_requirements(self, requirement: Requirement):
        assert isinstance(requirement, Requirement)
        if not isinstance(requirement.uid, str):
            return []

        if not requirement.uid or len(requirement.uid) == 0:
            return []

        if not self.requirements_parents:
            return []

        parent_requirements = self.requirements_parents[requirement.uid]['parents']
        return parent_requirements

    def has_parent_requirements(self, requirement: Requirement):
        assert isinstance(requirement, Requirement)
        if not isinstance(requirement.uid, str):
            return False

        if not requirement.uid or len(requirement.uid) == 0:
            return False

        if len(self.requirements_parents) == 0:
            return False

        parent_requirements = self.requirements_parents[requirement.uid]['parents']
        return len(parent_requirements) > 0

    def has_children_requirements(self, requirement: Requirement):
        assert isinstance(requirement, Requirement)
        if not isinstance(requirement.uid, str):
            return False

        if not requirement.uid or len(requirement.uid) == 0:
            return False

        if not self.requirements_parents:
            return False

        children_requirements = self.requirements_parents[requirement.uid]['children']
        return len(children_requirements) > 0

    def get_children_requirements(self, requirement: Requirement):
        assert isinstance(requirement, Requirement)
        if not isinstance(requirement.uid, str):
            return []

        if len(requirement.uid) == 0:
            return []

        if not self.requirements_parents:
            return []

        children_requirements = self.requirements_parents[requirement.uid]['children']
        return children_requirements

    def get_link(self, requirement):
        document = self.requirements_parents[requirement.uid]['document']
        return "{} - Traceability.html#{}".format(document.name, requirement.uid)

    def get_deep_link(self, requirement):
        document = self.requirements_parents[requirement.uid]['document']
        return "{} - Traceability Deep.html#{}".format(document.name, requirement.uid)

    def has_tags(self, document):
        if document.name not in self.tags_map:
            return False
        tags_bag = self.tags_map[document.name]
        return len(tags_bag.keys())

    def get_tags(self, document):
        assert document.name in self.tags_map
        tags_bag = self.tags_map[document.name]
        if not tags_bag:
            return []

        tags = sorted(tags_bag.keys(), key=alphanumeric_sort)
        for tag in tags:
            yield tag, tags_bag[tag]

    def get_max_ref_depth(self, document):
        return self.documents_ref_depth_map[document]



import collections

from strictdoc.backend.dsl.models.document import Document
from strictdoc.backend.dsl.models.requirement import CompositeRequirement, Requirement
from strictdoc.backend.dsl.models.section import FreeText, Section
from strictdoc.core.level_counter import LevelCounter


class DocumentCachingIterator:
    def __init__(self, document):
        assert isinstance(document, Document)

        self.document = document
        self.nodes_cache = []
        self.toc_nodes_cache = []

    def table_of_contents(self):
        if len(self.toc_nodes_cache) > 0:
            yield from self.toc_nodes_cache
            return

        for node in self.all_content():
            if isinstance(node, FreeText):
                continue

            self.toc_nodes_cache.append(node)
            yield node

    def all_content(self):
        if len(self.nodes_cache) > 0:
            yield from self.nodes_cache
            return

        document = self.document
        level_counter = LevelCounter()

        task_list = collections.deque(document.section_contents)

        while True:
            if not task_list:
                break

            current = task_list.popleft()

            if (isinstance(current, Section) or
                isinstance(current, CompositeRequirement) or
                isinstance(current, Requirement)):
                level_counter.adjust(current.ng_level)

                current.context.title_number_string = level_counter.get_string()

            self.nodes_cache.append(current)
            yield current

            if isinstance(current, Section):
                task_list.extendleft(reversed(current.section_contents))

            if isinstance(current, CompositeRequirement):
                task_list.extendleft(reversed(current.requirements))



import os
import sys
from multiprocessing import freeze_support

STRICTDOC_ROOT_PATH = os.path.join(os.path.dirname(__file__), "..", "..")
sys.path.append(STRICTDOC_ROOT_PATH)

from strictdoc.helpers.parallelizer import Parallelizer
from strictdoc.cli.cli_arg_parser import cli_args_parser
from strictdoc.core.actions.export_action import ExportAction
from strictdoc.core.actions.passthrough_action import PassthroughAction


def main():
    # Initializing the parallelizer early in the startup process because otherwise
    # fork() used by Python's multiprocessing library results in the following error:
    #
    # """
    # objc[7522]: +[__NSPlaceholderDictionary initialize] may have been in progress
    # in another thread when fork() was called. We cannot safely call it or ignore
    # it in the fork() child process. Crashing instead. Set a breakpoint on
    # objc_initializeAfterForkError to debug.
    # """
    enable_parallelization = '--no-parallelization' not in sys.argv
    PARALLELIZER = Parallelizer.create(enable_parallelization)


    parser = cli_args_parser()
    args = parser.parse_args()

    if args.command == 'passthrough':
        input_file = args.input_file
        if not os.path.isfile(input_file):
            sys.stdout.flush()
            err = "error: passthrough command's input file does not exist: {}".format(input_file)
            print(err)
            exit(1)

        output_file = args.output_file
        if output_file:
            output_dir = os.path.dirname(output_file)
            if not os.path.isdir(output_dir):
                print("not a directory: {}".format(output_file))
                exit(1)

        passthrough_action = PassthroughAction()
        passthrough_action.passthrough(input_file, output_file)

    elif args.command == 'export':
        parallelization_value = 'Enabled' if enable_parallelization else 'Disabled'
        print("Parallelization: {}".format(parallelization_value), flush=True)
        export_controller = ExportAction(STRICTDOC_ROOT_PATH, PARALLELIZER)
        export_controller.export(args.input_paths, args.output_dir)

    else:
        raise NotImplementedError


if __name__ == '__main__':
    main()



from jinja2 import Environment, PackageLoader


class DocumentTableHTMLGenerator:
    env = Environment(
        loader=PackageLoader('strictdoc', 'export/html/templates')
    )
    env.globals.update(isinstance=isinstance)

    @staticmethod
    def export(document, traceability_index, markup_renderer, link_renderer):
        output = ""

        template = DocumentTableHTMLGenerator.env.get_template(
            'single_document_table/document.jinja.html'
        )

        output += template.render(document=document,
                                  traceability_index=traceability_index,
                                  renderer=markup_renderer,
                                  link_renderer=link_renderer)

        return output



import glob
import os
from functools import partial
from pathlib import Path

from strictdoc.core.document_finder import DocumentFinder
from strictdoc.core.document_meta import DocumentMeta
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.generators.document import DocumentHTMLGenerator
from strictdoc.export.html.generators.document_table import DocumentTableHTMLGenerator
from strictdoc.export.html.generators.document_deep_trace import DocumentDeepTraceHTMLGenerator
from strictdoc.export.html.generators.document_trace import DocumentTraceHTMLGenerator
from strictdoc.export.html.generators.document_tree import DocumentTreeHTMLGenerator
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.export.html.renderers.markup_renderer import MarkupRenderer
from strictdoc.export.rst.export import SingleDocumentRSTExport
from strictdoc.helpers.file_modification_time import get_file_modification_time
from strictdoc.helpers.file_system import sync_dir
from strictdoc.helpers.timing import timing_decorator, measure_performance


class ExportAction:
    def __init__(self, strictdoc_src_path, parallelizer):
        assert parallelizer
        self.strictdoc_src_path = strictdoc_src_path
        self.cwd = os.getcwd()
        self.parallelizer = parallelizer
        strict_own_files = glob.iglob('{}/strictdoc/**/*'.format(self.strictdoc_src_path), recursive=True)
        strict_own_files = [f for f in strict_own_files if f.endswith('.html') or f.endswith('.py')]
        latest_strictdoc_own_file = max(strict_own_files, key=os.path.getctime)
        self.strictdoc_last_update = get_file_modification_time(latest_strictdoc_own_file)

    @timing_decorator('Export')
    def export(self, path_to_single_file_or_doc_root, output_dir):
        if isinstance(path_to_single_file_or_doc_root, str):
            path_to_single_file_or_doc_root = [path_to_single_file_or_doc_root]
        output_dir = output_dir if output_dir else "output"

        if not os.path.isabs(output_dir):
            output_dir = os.path.join(self.cwd, output_dir)

        output_html_root = '{}/html'.format(output_dir)
        output_rst_root = '{}/rst'.format(output_dir)
        output_html_static_files = '{}/_static'.format(output_html_root)

        Path(output_html_root).mkdir(parents=True, exist_ok=True)

        document_tree, asset_dirs = DocumentFinder.find_sdoc_content(
            path_to_single_file_or_doc_root, output_html_root, self.parallelizer
        )

        traceability_index = TraceabilityIndex.create(document_tree)
        markup_renderer = MarkupRenderer()
        link_renderer = LinkRenderer(output_html_root)

        writer = DocumentTreeHTMLGenerator()
        output = writer.export(document_tree)

        output_file = '{}/index.html'.format(output_html_root)

        with open(output_file, 'w') as file:
            file.write(output)

        # Single Document pages (RST)
        Path(output_rst_root).mkdir(parents=True, exist_ok=True)
        for document in document_tree.document_list:
            document_content = SingleDocumentRSTExport.export(document_tree,
                                                              document,
                                                              traceability_index)

            document_out_file = "{}/{}.rst".format(output_rst_root, document.name)
            with open(document_out_file, 'w') as file:
                file.write(document_content)

            # HACK:
            # ProcessPoolExecutor doesn't work because of non-picklable parts
            # of textx. The offending fields are stripped down because they
            # are not used anyway.
            document._tx_parser = None
            document._tx_attrs = None
            document._tx_metamodel = None
            document._tx_peg_rule = None

        export_binding = partial(ExportAction._export_with_performance,
                                 document_tree=document_tree,
                                 traceability_index=traceability_index,
                                 markup_renderer=markup_renderer,
                                 link_renderer=link_renderer,
                                 strictdoc_src_path=self.strictdoc_src_path,
                                 strictdoc_last_update=self.strictdoc_last_update)

        self.parallelizer.map(document_tree.document_list, export_binding)

        static_files_src = os.path.join(self.strictdoc_src_path,
                                        'strictdoc/export/html/static')
        sync_dir(static_files_src, output_html_static_files)
        for asset_dir in asset_dirs:
            source_path = asset_dir['full_path']
            output_relative_path = asset_dir['relative_path']
            destination_path = os.path.join(output_html_root, output_relative_path)
            sync_dir(source_path, destination_path)

        print('Export completed. Documentation tree can be found at:\n{}'.format(output_html_root))

    @staticmethod
    def _export_with_performance(document,
                                 document_tree,
                                 traceability_index,
                                 markup_renderer,
                                 link_renderer,
                                 strictdoc_src_path,
                                 strictdoc_last_update):
        document_meta: DocumentMeta = document.meta
        full_output_path = os.path.join(strictdoc_src_path, document_meta.get_html_doc_path())

        # If file exists we want to check its modification path in order to skip
        # its generation in case it has not changed since the last generation.
        if os.path.isfile(full_output_path):
            output_file_mtime = get_file_modification_time(full_output_path)
            sdoc_mtime = get_file_modification_time(document_meta.input_doc_full_path)

            if (sdoc_mtime < output_file_mtime and
                strictdoc_last_update < output_file_mtime):
                with measure_performance('Skip: {}'.format(document.name)):
                    return

        with measure_performance('Published: {}'.format(document.name)):
            ExportAction._export(document, document_tree, traceability_index,
                                 markup_renderer,
                                 link_renderer)
        return None

    @staticmethod
    def _export(document, document_tree, traceability_index,
                markup_renderer,
                link_renderer):
        document_meta: DocumentMeta = document.meta

        document_output_folder = document_meta.output_document_dir_full_path
        Path(document_output_folder).mkdir(parents=True, exist_ok=True)

        # Single Document pages
        document_content = DocumentHTMLGenerator.export(document_tree,
                                                        document,
                                                        traceability_index,
                                                        markup_renderer,
                                                        link_renderer)

        document_out_file = document_meta.get_html_doc_path()
        with open(document_out_file, 'w') as file:
            file.write(document_content)

        # Single Document Table pages
        document_content = DocumentTableHTMLGenerator.export(
            document, traceability_index, markup_renderer, link_renderer
        )
        document_out_file = document_meta.get_html_table_path()

        with open(document_out_file, 'w') as file:
            file.write(document_content)

        # Single Document Traceability pages
        document_content = DocumentTraceHTMLGenerator.export(
            document, traceability_index, markup_renderer, link_renderer
        )
        document_out_file = document_meta.get_html_traceability_path()

        with open(document_out_file, 'w') as file:
            file.write(document_content)

        # Single Document Deep Traceability pages
        document_content = DocumentDeepTraceHTMLGenerator.export_deep(
            document, traceability_index, markup_renderer, link_renderer
        )
        document_out_file = document_meta.get_html_deep_traceability_path()

        with open(document_out_file, 'w') as file:
            file.write(document_content)

        return document




class LevelCounter:
    def __init__(self):
        self.levels = []

    def adjust(self, level):
        current_level = len(self.levels)

        if current_level < level:
            self.levels.append(1)
        elif current_level > level:
            while len(self.levels) > level:
                self.levels.pop()

            self.levels[-1] += 1
        else:
            self.levels[-1] += 1

    def get_string(self):
        str_levels = map(lambda el: str(el), self.levels)
        string = '.'.join(str_levels)

        return string



import os


class FileOrFolderEntry:
    def get_full_path(self):
        raise NotImplementedError

    def get_level(self):
        raise NotImplementedError

    def is_folder(self):
        raise NotImplementedError


class File(FileOrFolderEntry):
    def __init__(self, level, full_path):
        assert os.path.isfile(full_path)
        self.level = level
        self.full_path = full_path
        self.root_path = full_path
        self.files = [self]
        self.subfolder_trees = []

    def __repr__(self):
        return "File: {}".format(self.full_path)

    def is_folder(self):
        return False

    def get_full_path(self):
        return self.full_path

    def get_level(self):
        return self.level

    def get_file_name(self):
        return os.path.basename(self.full_path)


class FileTree(FileOrFolderEntry):
    def __init__(self, root_path, level):
        assert os.path.isdir(root_path)

        self.root_path = root_path
        self.level = level
        self.files = []
        self.subfolder_trees = []

    def __repr__(self):
        return "FileTree: (root_path: {}, files: {})".format(self.root_path, self.files)

    def is_folder(self):
        return True

    def get_full_path(self):
        return self.root_path

    def get_level(self):
        return self.level

    def get_folder_name(self):
        return os.path.basename(os.path.normpath(self.root_path))

    def set(self, files):
        for file in files:
            full_file_path = os.path.join(self.root_path, file)
            self.files.append(File(self.level + 1, full_file_path))

    def add_subfolder_tree(self, subfolder_tree):
        assert isinstance(subfolder_tree, FileTree)
        self.subfolder_trees.append(subfolder_tree)

    def sort_subfolder_trees(self):
        self.subfolder_trees.sort(key=lambda subfolder: subfolder.root_path)

    def dump(self):
        print(self)
        for subfolder in self.subfolder_trees:
            subfolder.dump()


class DocumentTree:
    def __init__(self,
                 file_tree,
                 document_list,
                 map_docs_by_paths):
        assert isinstance(file_tree, list)
        assert isinstance(document_list, list)
        assert isinstance(map_docs_by_paths, dict)
        self.file_tree = file_tree
        self.document_list = document_list
        self.map_docs_by_paths = map_docs_by_paths

    def __repr__(self):
        return "DocumentTree: {} document_list: {}".format(self.file_tree, self.document_list)

    def get_document_by_path(self, doc_full_path):
        document = self.map_docs_by_paths[doc_full_path]
        return document



import os
import sys

from strictdoc.backend.dsl.models.document import Document
from strictdoc.backend.dsl.reader import SDReader
from strictdoc.core.document_meta import DocumentMeta
from strictdoc.core.document_tree import FileTree, DocumentTree, File
from strictdoc.helpers.sorting import alphanumeric_sort
from strictdoc.helpers.timing import measure_performance, timing_decorator


class DocumentFinder:
    @staticmethod
    @timing_decorator('Find')
    def find_sdoc_content(paths_to_files_or_docs, output_root_html, parallelizer):
        for paths_to_files_or_doc in paths_to_files_or_docs:
            if not os.path.exists(paths_to_files_or_doc):
                sys.stdout.flush()
                err = "error: Provided path is neither a single file or a folder: '{}'".format(
                    paths_to_files_or_doc
                )
                print(err)
                exit(1)

        file_tree, asset_dirs = DocumentFinder._build_file_tree(paths_to_files_or_docs)
        document_tree = DocumentFinder._build_document_tree(file_tree,
                                                            output_root_html,
                                                            parallelizer)

        return document_tree, asset_dirs

    @staticmethod
    def _iterate_file_trees(file_trees):
        for file_tree in file_trees:
            task_list = [file_tree]

            file_tree_mount_folder = os.path.basename(file_tree.root_path)

            while len(task_list) > 0:
                current_tree = task_list.pop(0)

                for doc_file in current_tree.files:
                    yield file_tree, doc_file, file_tree_mount_folder

                task_list.extend(current_tree.subfolder_trees)

    @staticmethod
    def processing(document_triple):
        file_tree, doc_file, file_tree_mount_folder = document_triple
        doc_full_path = doc_file.get_full_path()

        with measure_performance('Reading SDOC: {}'.format(os.path.basename(doc_full_path))):
            reader = SDReader()
            document = reader.read_from_file(doc_full_path)
            assert isinstance(document, Document)

        document._tx_parser = None
        document._tx_attrs = None
        document._tx_metamodel = None
        document._tx_peg_rule = None
        document._tx_model_params = None
        return doc_file, document

    @staticmethod
    def _build_document_tree(file_trees, output_root_html, parallelizer):
        assert isinstance(file_trees, list)
        document_list, map_docs_by_paths, map_relpaths_by_docs = [], {}, {}

        file_tree_list = list(DocumentFinder._iterate_file_trees(file_trees))
        found_documents = parallelizer.map(
            file_tree_list,
            DocumentFinder.processing
        )

        for doc_file, document in found_documents:
            input_doc_full_path = doc_file.get_full_path()
            map_docs_by_paths[input_doc_full_path] = document
            document_list.append(document)

        for file_tree, doc_file, file_tree_mount_folder in file_tree_list:
            input_doc_full_path = doc_file.get_full_path()
            document = map_docs_by_paths[input_doc_full_path]
            assert isinstance(document, Document)

            doc_relative_path = os.path.relpath(input_doc_full_path, file_tree.root_path)
            doc_relative_path_folder = os.path.dirname(doc_relative_path)

            output_document_dir_rel_path = '{}/{}'.format(
                file_tree_mount_folder,
                doc_relative_path_folder
            ) if doc_relative_path_folder else file_tree_mount_folder

            document_filename = os.path.basename(input_doc_full_path)
            document_filename_base = os.path.splitext(document_filename)[0]

            output_document_dir_full_path = '{}/{}'.format(
                output_root_html, output_document_dir_rel_path
            )

            document_meta = DocumentMeta(doc_file.level,
                                         document_filename_base,
                                         input_doc_full_path,
                                         output_document_dir_full_path,
                                         output_document_dir_rel_path)

            document.assign_meta(document_meta)

            map_docs_by_paths[input_doc_full_path] = document

        return DocumentTree(file_trees,
                            document_list,
                            map_docs_by_paths)

    @staticmethod
    def _build_file_tree(paths_to_files_or_docs):
        asset_dirs = []
        root_trees = []

        for path_to_doc_root_raw in paths_to_files_or_docs:
            if os.path.isfile(path_to_doc_root_raw):
                root_trees.append(File(0, path_to_doc_root_raw))
                continue

            # Strip away the trailing slash to let the later os.path.relpath
            # calculations work correctly.
            path_to_doc_root = path_to_doc_root_raw.rstrip('/')
            path_to_doc_root_base = os.path.dirname(path_to_doc_root)
            root_level = path_to_doc_root.count(os.sep)

            tree_map = {path_to_doc_root: FileTree(path_to_doc_root, 0)}

            for current_root_path, dirs, files in os.walk(path_to_doc_root, topdown=False):
                if os.path.basename(current_root_path) == '_assets':
                    asset_dirs.append({
                        'full_path': current_root_path,
                        'relative_path': os.path.relpath(current_root_path, path_to_doc_root_base)
                    })

                current_root_path_level = current_root_path.count(os.sep) - root_level

                if current_root_path not in tree_map:
                    tree_map[current_root_path] = FileTree(current_root_path, current_root_path_level)
                current_tree = tree_map[current_root_path]

                current_tree.sort_subfolder_trees()

                files = [f for f in files if f.endswith('.sdoc')]
                if len(files) > 0:
                    files.sort(key=alphanumeric_sort)
                    current_tree.set(files)

                    if current_root_path == path_to_doc_root:
                        continue

                    current_parent_path = os.path.dirname(current_root_path)
                    if current_parent_path not in tree_map:
                        tree_map[current_parent_path] = FileTree(current_parent_path, current_root_path_level - 1)
                    current_parent_tree = tree_map[current_parent_path]
                    current_parent_tree.add_subfolder_tree(current_tree)

            root_trees.append(tree_map[path_to_doc_root])
        return root_trees, asset_dirs



import collections

from strictdoc.core.document_tree import FileTree, DocumentTree


class DocumentTreeIterator:
    def __init__(self, document_tree):
        assert isinstance(document_tree, DocumentTree)
        self.document_tree = document_tree

    def iterator(self):
        task_list = collections.deque(self.document_tree.file_tree)

        while task_list:
            file_tree_or_file = task_list.popleft()
            yield file_tree_or_file
            if isinstance(file_tree_or_file, FileTree):
                task_list.extendleft(reversed(file_tree_or_file.files))
                task_list.extendleft(reversed(file_tree_or_file.subfolder_trees))



import string


# Strings with embedded variables in Python
# https://stackoverflow.com/a/16553401/598057
class RubyTemplate(string.Template):
    delimiter = '#'


REQUIREMENT_FIELDS = """
  ('UID: ' uid = /.+$/ '\n')? 

  ('STATUS: ' status = RequirementStatus '\n')?

  ('TAGS: ' (tags = TagRegex) tags *= TagXs '\n')?

  ('REFS:' '\n' references *= Reference)?

  ('TITLE: ' title = /.*$/ '\n')?

  ('STATEMENT: ' (statement = SingleLineString | statement_multiline = MultiLineString) '\n')?

  ('BODY: ' body = MultiLineString '\n' )?

  comments *= RequirementComment
"""

STRICTDOC_GRAMMAR = RubyTemplate("""
Document[noskipws]:
  '[DOCUMENT]' '\n'
  'NAME: ' name = /.*$/ '\n'
  free_texts *= SpaceThenFreeText
  section_contents *= SectionOrRequirement
;

Section[noskipws]:
  '[SECTION]' 
  '\n'
  'LEVEL: ' level = /[1-6]/ '\n'
  'TITLE: ' title = /.*$/ '\n'
  free_texts *= SpaceThenFreeText
  section_contents *= SectionOrRequirement
  '\n'
  '[/SECTION]'
  '\n'
;

SectionOrRequirement[noskipws]:
  '\n' (Section | Requirement | CompositeRequirement)
;

SpaceThenRequirement[noskipws]:
  '\n' (Requirement | CompositeRequirement)
;

SpaceThenFreeText[noskipws]:
  '\n' (FreeText)
;

Requirement[noskipws]:
  '[REQUIREMENT]' '\n'

  #{REQUIREMENT_FIELDS}
;

CompositeRequirement[noskipws]:
  '[COMPOSITE-REQUIREMENT]' '\n'

  #{REQUIREMENT_FIELDS}
  
  requirements *= SpaceThenRequirement

  '\n'
  '[/COMPOSITE-REQUIREMENT]' '\n'
;

TagRegex[noskipws]:
  /[\w\/-]+( *[\w\/-]+)*/
;

TagXs[noskipws]:
  /, /- TagRegex
;

RequirementStatus[noskipws]:
  'Draft' | 'Active' | 'Deleted';

RequirementComment[noskipws]:
  'COMMENT: ' (comment_single = SingleLineString | comment_multiline = MultiLineString) '\n'
;

SingleLineString:
  /[^>]{3}.*$/
;

MultiLineString:
  /(?ms)>>>\n(.*?)\n<<</
;

Reference[noskipws]:
  '- TYPE: ' ref_type = ReferenceType '\n'
  '  VALUE: ' path = /.*$/ '\n'
;

ReferenceType[noskipws]:
  'File' | 'Parent'
;

FreeText[noskipws]:
  text = /(?ms)\[FREETEXT\]\n(.*?)\n\[\/FREETEXT\]/ '\n'
;
""").substitute(REQUIREMENT_FIELDS=REQUIREMENT_FIELDS)



import argparse


def cli_args_parser():
    # for arg in sys.argv:
    #     if arg == '--help':
    #         # print_help()
    #         assert 0
    #         exit(0)
    #
    # for arg in sys.argv:
    #     if arg == '--version':
    #         # print_version()
    #         assert 0
    #         exit(0)

    # if os.path.getsize(check_file) == 0:
    #     sys.stdout.flush()
    #     print("error: no check strings found with prefix 'CHECK:'", file=sys.stderr)
    #     exit(2)

    # https://stackoverflow.com/a/19476216/598057
    main_parser = argparse.ArgumentParser()

    command_subparsers = main_parser.add_subparsers(title="command",
                                                    dest="command")
    command_subparsers.required = True

    command_parser_export = command_subparsers.add_parser(
        "export",
        help="Export document tree.",
        parents=[],
        description="Export command: input SDoc documents are generated into HTML and other formats."
    )
    command_parser_export.add_argument('input_paths',
                                        type=str,
                                        nargs='+',
                                        help='One or more folders with *.sdoc files')
    command_parser_export.add_argument('--output-dir',
                                       type=str,
                                       help='Output folder')
    command_parser_export.add_argument('--no-parallelization',
                                       action='store_true',
                                       help='Disables parallelization. All work happens in the main thread. This option might be useful for debugging.')

    command_parser_passthrough = command_subparsers.add_parser(
        "passthrough", help="Export document tree.", parents=[]
    )
    command_parser_passthrough.add_argument('input_file',
                                            type=str,
                                            help='Path to the input SDoc file')

    command_parser_passthrough.add_argument('--output-file',
                                            type=str,
                                            help='Path to the output SDoc file')

    return main_parser



import contextlib
from functools import wraps
from time import time


def timing_decorator(name):
    def timing_internal(f):
        @wraps(f)
        def wrap(*args, **kw):
            ts = time()
            result = f(*args, **kw)
            te = time()
            print("Step '{}' took: {} sec".format(name, round(te-ts, 2)))
            return result
        return wrap
    return timing_internal


@contextlib.contextmanager
def measure_performance(title):
    ts = time()
    yield
    te = time()

    padded_name = '{title} '.format(title=title).ljust(60, '.')
    padded_time = ' {:0.2f}'.format((te-ts)).rjust(6, '.')
    print('{}{}s'.format(padded_name, padded_time))



class Reference(object):
    def __init__(self, parent, ref_type, path):
        self.parent = parent
        self.ref_type = ref_type
        self.path = path.strip()

    def __str__(self):
        return "Reference: <ref_type = {}, path = {}>".format(
            self.ref_type, self.path
        )

    def __repr__(self):
        return self.__str__()



from strictdoc.backend.dsl.models.reference import Reference


class RequirementContext(object):
    def __init__(self):
        self.title_number_string = None


class Requirement(object):
    def __init__(self,
                 parent,
                 statement,
                 statement_multiline,
                 uid,
                 status,
                 tags,
                 references,
                 title,
                 body,
                 comments,
                 requirements=None):
        assert parent

        self.parent = parent

        # TODO: Why textX creates empty uid when the sdoc doesn't declare the
        # UID field?
        self.uid = uid.strip() if (isinstance(uid, str) and len(uid) > 0) else None
        self.status = status
        self.tags = tags
        self.references: [Reference] = references
        self.title = title
        self.statement = statement
        self.statement_multiline = statement_multiline
        self.body = body
        self.comments = comments
        self.requirements = requirements

        # TODO: Is it worth to move this to dedicated Presenter* classes to
        # keep this class textx-only?
        self.ng_level = None
        self.context = RequirementContext()

    def __str__(self):
        return "{}: <ng_level: {}, uid: {}, title_or_none: {}, statement: {}>".format(
            self.__class__.__name__,
            self.ng_level,
            self.uid, self.title, self.statement
        )

    def __repr__(self):
        return self.__str__()

    @property
    def has_meta(self):
        return (
            self.uid is not None or
            (self.tags is not None and len(self.tags) > 0) or
            self.status is not None
        )

    @property
    def is_requirement(self):
        return True

    @property
    def is_section(self):
        return False

    @property
    def is_composite_requirement(self):
        return False

    def get_statement_single_or_multiline(self):
        if self.statement:
            return self.statement
        elif self.statement_multiline:
            return self.statement_multiline
        raise NotImplementedError


class CompositeRequirement(Requirement):
    def __init__(self, parent, **fields):
        super(CompositeRequirement, self).__init__(parent, **fields)
        self.ng_sections = []

    @property
    def is_composite_requirement(self):
        return True


class Body(object):
    def __init__(self, parent, content):
        self.parent = parent
        self.content = content.strip()

    def __str__(self):
        return "Body: <{}>".format(
            self.content
        )

    def __repr__(self):
        return self.__str__()


class RequirementComment(object):
    def __init__(self, parent, comment_single, comment_multiline):
        self.parent = parent
        self.comment_single = comment_single
        self.comment_multiline = comment_multiline

    def __str__(self):
        return "Comment: <>".format()

    def __repr__(self):
        return self.__str__()

    def get_comment(self):
        comment = self.comment_single if self.comment_single else self.comment_multiline
        return comment



import os

from strictdoc.core.document_tree import DocumentTree
from strictdoc.export.rst.writer import RSTWriter


def get_path_components(folder_path):
    path = os.path.normpath(folder_path)
    return path.split(os.sep)


class SingleDocumentRSTExport:
    @staticmethod
    def export(document_tree: DocumentTree, document, traceability_index):
        writer = RSTWriter()

        single_or_many = len(document_tree.document_list) == 1
        output = writer.write(document, single_or_many)

        return output



import multiprocessing
import sys


# This might be not the most efficient way to parallelize using processes,
# however it does not produce the following problem:
# objc[24250]: +[__NSPlaceholderDate initialize] may have been in progress in another thread when fork() was called.
# The suggestion solution to this problem is to use
# OBJC_DISABLE_INITIALIZE_FORK_SAFETY=YES
# which seems to be a worth hack than using a little bit slower parallelization
# algorithm below (slower means roughly 7s instead of 6s, so 6/7).
class Parallelizer:
    @staticmethod
    def create(parallelize):
        if parallelize:
            return Parallelizer()
        else:
            return NullParallelizer()

    def __init__(self):
        self.input_queue = multiprocessing.Queue()
        self.output_queue = multiprocessing.Queue()
        self.pool = multiprocessing.Pool(multiprocessing.cpu_count(),
                                         Parallelizer._run,
                                         (self.input_queue, self.output_queue,))

    def map(self, contents, processing_func):
        size = 0
        for content in contents:
            self.input_queue.put((content, processing_func))
            size += 1
        results = []
        while size > 0:
            result = self.output_queue.get(block=True)
            results.append(result)
            size -= 1
        return results

    @staticmethod
    def _run(input_queue, output_queue):
        while True:
            content, processing_func = input_queue.get(block=True)
            result = processing_func(content)
            sys.stdout.flush()
            sys.stderr.flush()
            output_queue.put(result)


class NullParallelizer:
    @staticmethod
    def map(contents, processing_func):
        results = []
        for content in contents:
            results.append(processing_func(content))
        return results



import traceback

from textx import metamodel_from_str

from strictdoc.backend.dsl.grammar import STRICTDOC_GRAMMAR
from strictdoc.backend.dsl.models.document import Document
from strictdoc.backend.dsl.models.reference import Reference
from strictdoc.backend.dsl.models.requirement import (
    Requirement, CompositeRequirement, RequirementComment, Body
)
from strictdoc.backend.dsl.models.section import Section, FreeText

DOCUMENT_MODELS = [
    Document,
    RequirementComment,
    Section,
    Requirement,
    CompositeRequirement,
    # Body,
    Reference,
    FreeText
]


def section_obj_processor(section):
    section.parent.ng_sections.append(section)


# During parsing, it is often the case that the node's parents do not have their
# levels resolved yet. We go up the parent chain and resolve all of the parents
# manually.
def resolve_parents(node):
    parents_to_resolve_level = []
    cursor = node.parent
    while cursor.ng_level is None:
        parents_to_resolve_level.append(cursor)
        cursor = cursor.parent
    for parent_idx, parent in enumerate(reversed(parents_to_resolve_level)):
        parent.ng_level = cursor.ng_level + parent_idx + 1


def composite_requirement_obj_processor(composite_requirement):
    if isinstance(composite_requirement.parent, Section):
        if not composite_requirement.parent.ng_level:
            resolve_parents(composite_requirement)
        assert composite_requirement.parent.level
        composite_requirement.ng_level = composite_requirement.parent.level + 1

        composite_requirement.parent.ng_sections.append(composite_requirement)
    elif isinstance(composite_requirement.parent, CompositeRequirement):
        if not composite_requirement.parent.ng_level:
            resolve_parents(composite_requirement)
        assert composite_requirement.parent.ng_level
        composite_requirement.ng_level = composite_requirement.parent.ng_level + 1
    elif isinstance(composite_requirement.parent, Document):
        composite_requirement.ng_level = 1
    else:
        raise NotImplementedError


def requirement_obj_processor(requirement):
    if isinstance(requirement.parent, Section):
        assert requirement.parent.level
        requirement.ng_level = requirement.parent.level + 1
    elif isinstance(requirement.parent, CompositeRequirement):
        if not requirement.parent.ng_level:
            resolve_parents(requirement)
        requirement.ng_level = requirement.parent.ng_level + 1
    elif isinstance(requirement.parent, Document):
        requirement.ng_level = 1
    else:
        raise NotImplementedError

def freetext_obj_processor(free_text):
    if isinstance(free_text.parent, Section):
        assert free_text.parent.level
        free_text.ng_level = free_text.parent.level + 1
    elif isinstance(free_text.parent, CompositeRequirement):
        if not free_text.parent.ng_level:
            resolve_parents(free_text)
        free_text.ng_level = free_text.parent.ng_level + 1
    elif isinstance(free_text.parent, Document):
        free_text.ng_level = 1
    else:
        raise NotImplementedError

class SDReader:
    def __init__(self):
        self.meta_model = metamodel_from_str(STRICTDOC_GRAMMAR,
                                             classes=DOCUMENT_MODELS,
                                             use_regexp_group=True)
        obj_processors = {
            'Section': section_obj_processor,
            'CompositeRequirement': composite_requirement_obj_processor,
            'Requirement': requirement_obj_processor,
            'FreeText': freetext_obj_processor
        }

        self.meta_model.register_obj_processors(obj_processors)

    def read(self, input):
        document = self.meta_model.model_from_str(input)
        return document

    def read_from_file(self, file_path):
        with open(file_path, 'r') as file:
            sdoc_content = file.read()

        try:
            sdoc = self.read(sdoc_content)
            return sdoc
        except NotImplementedError as exc:
            traceback.print_exc()
            exit(1)
        except Exception as exc:
            print("error: could not parse file: {}.\n{}: {}".format(file_path, exc.__class__.__name__,  exc))
            traceback.print_exc()
            exit(1)



from enum import Enum

from strictdoc.backend.dsl.models.requirement import Requirement
from strictdoc.backend.dsl.models.section import FreeText, Section
from strictdoc.core.document_iterator import DocumentCachingIterator


class TAG(Enum):
    SECTION = 1
    REQUIREMENT = 2
    COMPOSITE_REQUIREMENT = 3


class RSTWriter:
    def __init__(self):
        pass

    def write(self, document, single_document):
        document_iterator = DocumentCachingIterator(document)
        output = ""

        if not single_document:
            output += self._print_rst_header(document.name, 0)

        for free_text in document.free_texts:
            output += self._print_free_text(free_text)

        for content_node in document_iterator.all_content():
            if isinstance(content_node, Section):
                output += self._print_rst_header(content_node.title, content_node.ng_level)
                for free_text in content_node.free_texts:
                    output += self._print_free_text(free_text)

            elif isinstance(content_node, Requirement):
                output += self._print_rst_header(content_node.title, content_node.ng_level)
                output += self._print_requirement_fields(content_node)

        return output.lstrip()

    def _print_rst_header(self, string, level):
        chars = {
            0: '$',
            1: '=',
            2: '-',
            3: '~',
            4: '^',
            5: '"',
            6: '#',
            7: "'",
        }
        header_char = chars[level]
        output = ''
        output += string
        output += "\n"
        output += header_char.rjust(len(string), header_char)
        output += "\n"
        output += "\n"
        return output

    def _print_requirement_fields(self, section_content):
        output = ''

        if section_content.uid:
            output += "``["
            output += section_content.uid
            output += "]``"
            output += "\n"
            output += "\n"

        if section_content.statement:
            output += section_content.statement
            output += "\n"
            output += "\n"
        elif section_content.statement_multiline:
            output += section_content.statement_multiline
            output += "\n"
            output += "\n"
        else:
            pass  # raise RuntimeError('Statement is missing')

        if section_content.body:
            output += "\n"
            output += section_content.body
            output += "\n"
            output += "\n"

        for comment in section_content.comments:
            output += '**'
            output += "Comment:"
            output += '** '
            output += comment.get_comment()
            output += "\n"
            output += "\n"

        return output

    def _print_free_text(self, free_text):
        assert isinstance(free_text, FreeText)
        output = ''
        output += free_text.text
        output += '\n'
        output += '\n'

        return output



import distutils.log
import distutils.dir_util
import os


def sync_dir(src_dir, dst_dir):
    assert os.path.isdir(src_dir)
    distutils.log.set_verbosity(distutils.log.DEBUG)
    distutils.dir_util.copy_tree(
        src_dir,
        dst_dir,
        update=1,
        verbose=1,
    )



from jinja2 import Environment, PackageLoader


class DocumentHTMLGenerator:
    env = Environment(
        loader=PackageLoader('strictdoc', 'export/html/templates')
    )
    env.globals.update(isinstance=isinstance)

    @staticmethod
    def export(document_tree, document, traceability_index,
               markup_renderer,
               link_renderer):
        output = ""

        template = DocumentHTMLGenerator.env.get_template('single_document/document.jinja.html')

        output += template.render(document=document,
                                  traceability_index=traceability_index,
                                  link_renderer=link_renderer,
                                  renderer=markup_renderer)

        return output



class DocumentMeta:
    def __init__(self,
                 level,
                 document_filename_base,
                 input_doc_full_path,
                 output_document_dir_full_path,
                 output_document_dir_rel_path):
        self.level = level
        self.document_filename_base = document_filename_base
        self.input_doc_full_path = input_doc_full_path
        self.output_document_dir_full_path = output_document_dir_full_path
        self.output_document_dir_rel_path = output_document_dir_rel_path

    # Paths
    def get_html_doc_path(self):
        return "{}/{}.html".format(self.output_document_dir_full_path,
                                   self.document_filename_base)

    def get_html_table_path(self):
        return "{}/{} - TABLE.html".format(self.output_document_dir_full_path,
                                           self.document_filename_base)

    def get_html_traceability_path(self):
        return "{}/{} - TRACE.html".format(self.output_document_dir_full_path,
                                           self.document_filename_base)

    def get_html_deep_traceability_path(self):
        return "{}/{} - DEEP-TRACE.html".format(self.output_document_dir_full_path,
                                                self.document_filename_base)

    # Links
    def get_html_doc_link(self):
        return "{}/{}.html".format(self.output_document_dir_rel_path,
                                   self.document_filename_base)

    def get_html_table_link(self):
        return "{}/{} - TABLE.html".format(self.output_document_dir_rel_path,
                                           self.document_filename_base)

    def get_html_traceability_link(self):
        return "{}/{} - TRACE.html".format(self.output_document_dir_rel_path,
                                           self.document_filename_base)

    def get_html_deep_traceability_link(self):
        return "{}/{} - DEEP-TRACE.html".format(self.output_document_dir_rel_path,
                                                self.document_filename_base)

    def get_root_path_prefix(self):
        if self.level == 0:
            return '..'
        return ('../' * self.level)[:-1]



import inspect


class Logger:
    available_loggers = []
    enabled_loggers = []

    def __init__(self, logger_class):
        assert isinstance(logger_class, str)
        assert logger_class not in Logger.available_loggers
        Logger.available_loggers.append(logger_class)
        self.logger_class_name = logger_class

    def info(self, message):
        if self.logger_class_name not in self.enabled_loggers:
            return

        curframe = inspect.currentframe()
        calframe = inspect.getouterframes(curframe, 2)
        # print(curframe)
        # print(calframe[1])

        caller_file = calframe[1][1]
        caller_line = calframe[1][2]
        caller_name = calframe[1][3]

        print('{}.{}>\n"""\n{}\n"""\n{}:{}\n\n'.format(
            self.logger_class_name, caller_name,
            message,
            caller_file, caller_line
        ))



from jinja2 import Environment, PackageLoader

from strictdoc.core.document_tree_iterator import DocumentTreeIterator


class DocumentTreeHTMLGenerator:
    env = Environment(
        loader=PackageLoader('strictdoc', 'export/html/templates')
    )
    env.globals.update(isinstance=isinstance)

    @staticmethod
    def export(document_tree):
        document_tree_iterator = DocumentTreeIterator(document_tree)

        template = DocumentTreeHTMLGenerator.env.get_template('document_tree/document_tree.jinja.html')
        output = template.render(document_tree=document_tree,
                                 artefact_list=document_tree_iterator.iterator())

        return output



from enum import Enum

from strictdoc.backend.dsl.models.requirement import Requirement, CompositeRequirement
from strictdoc.backend.dsl.models.section import Section, FreeText
from strictdoc.core.document_iterator import DocumentCachingIterator


class TAG(Enum):
    SECTION = 1
    REQUIREMENT = 2
    COMPOSITE_REQUIREMENT = 3


class SDWriter:
    def __init__(self):
        pass

    def write(self, document):
        document_iterator = DocumentCachingIterator(document)
        output = ""

        output += "[DOCUMENT]"
        output += "\n"

        output += "NAME: "
        output += document.name
        output += "\n"

        for free_text in document.free_texts:
            output += "\n"
            output += self._print_free_text(free_text)

        closing_tags = []
        for content_node in document_iterator.all_content():
            while len(closing_tags) > 0 and content_node.ng_level <= closing_tags[-1][1]:
                closing_tag, level = closing_tags.pop()
                output += self._print_closing_tag(closing_tag)

            output += "\n"

            if isinstance(content_node, Section):
                output += self._print_section(content_node)
                closing_tags.append((TAG.SECTION, content_node.ng_level))
            elif isinstance(content_node, Requirement):
                if isinstance(content_node, CompositeRequirement):
                    output += "[COMPOSITE-REQUIREMENT]\n"
                    closing_tags.append((TAG.COMPOSITE_REQUIREMENT, content_node.ng_level))
                else:
                    output += "[REQUIREMENT]\n"

                output += self._print_requirement_fields(content_node)

        for closing_tag, level in reversed(closing_tags):
            output += self._print_closing_tag(closing_tag)

        return output

    def _print_section(self, section):
        assert isinstance(section, Section)
        output = ''
        output += "[SECTION]"
        output += "\n"
        output += "LEVEL: "
        output += str(section.level)
        output += "\n"
        output += "TITLE: "
        output += str(section.title)
        output += "\n"
        for free_text in section.free_texts:
            output += "\n"
            output += self._print_free_text(free_text)
        return output

    def _print_requirement_fields(self, section_content):
        output = ''

        if section_content.uid:
            output += "UID: "
            output += section_content.uid
            output += "\n"

        if section_content.status:
            output += "STATUS: "
            output += section_content.status
            output += "\n"

        if section_content.tags:
            output += "TAGS: "
            output += ', '.join(section_content.tags)
            output += '\n'

        if section_content.references:
            output += "REFS:"
            output += "\n"

            for reference in section_content.references:
                output += "- TYPE: "
                output += reference.ref_type
                output += "\n"
                output += "  VALUE: "
                output += reference.path
                output += "\n"

        if section_content.title:
            output += "TITLE: "
            output += section_content.title
            output += "\n"

        if section_content.statement:
            output += "STATEMENT: "
            output += section_content.statement
            output += "\n"
        elif section_content.statement_multiline:
            output += "STATEMENT: >>>"
            output += "\n"
            output += section_content.statement_multiline
            output += "\n"
            output += "<<<"
            output += "\n"
        else:
            raise RuntimeError('Statement is missing: {}'.format(section_content))

        if section_content.body:
            output += "BODY: >>>"
            output += "\n"
            output += section_content.body
            output += "\n"
            output += "<<<"
            output += "\n"

        for comment in section_content.comments:
            if comment.comment_multiline:
                output += "COMMENT: >>>\n"
                output += comment.comment_multiline
                output += "\n<<<\n"
            else:
                output += "COMMENT: "
                output += comment.comment_single
                output += "\n"

        return output

    def _print_closing_tag(self, closing_tag):
        output = ''
        if closing_tag == TAG.SECTION:
            output += '\n'
            output += '[/SECTION]'
            output += '\n'
        if closing_tag == TAG.COMPOSITE_REQUIREMENT:
            output += '\n'
            output += '[/COMPOSITE-REQUIREMENT]'
            output += '\n'
        return output

    def _print_free_text(self, free_text):
        assert isinstance(free_text, FreeText)
        output = ''
        output += '[FREETEXT]'
        output += '\n'
        output += free_text.text.rstrip()
        output += '\n'
        output += '[/FREETEXT]'
        output += '\n'
        return output



class SectionContext(object):
    def __init__(self):
        self.title_number_string = None


class Section(object):
    def __init__(self, parent, level, title, free_texts, section_contents):
        self.parent = parent
        self.level = int(level)
        self.title = title

        self.free_texts = free_texts
        self.section_contents = section_contents

        self.ng_level = self.level
        self.ng_sections = []
        self.context = SectionContext()

    def __str__(self):
        return "Section: <level: {}, title: {}, section_contents: {}>".format(
            self.level, self.title, self.section_contents
        )

    def __repr__(self):
        return self.__str__()

    @property
    def is_requirement(self):
        return False

    @property
    def is_section(self):
        return True


class FreeText:
    def __init__(self, parent, text):
        self.parent = parent
        self.text = text
        self.ng_level = None

    @property
    def is_requirement(self):
        return False

    @property
    def is_section(self):
        return False




from strictdoc.backend.dsl.models.requirement import Requirement, RequirementComment
from strictdoc.backend.dsl.models.section import FreeText
from strictdoc.export.rst.rst_to_html_fragment_writer import RstToHtmlFragmentWriter


class MarkupRenderer:
    def __init__(self):
        self.cache = {}

    def render_requirement_statement(self, requirement):
        assert isinstance(requirement, Requirement)
        if requirement in self.cache:
            return self.cache[requirement]

        output = RstToHtmlFragmentWriter.write(
            requirement.get_statement_single_or_multiline()
        )
        self.cache[requirement] = output

        return output

    def render_comment(self, comment):
        assert isinstance(comment, RequirementComment)
        if comment in self.cache:
            return self.cache[comment]

        output = RstToHtmlFragmentWriter.write(comment.get_comment())
        self.cache[comment] = output

        return output

    def render_free_text(self, free_text):
        assert isinstance(free_text, FreeText)
        if free_text in self.cache:
            return self.cache[free_text]

        output = RstToHtmlFragmentWriter.write(free_text.text)
        self.cache[free_text] = output

        return output



import re

from strictdoc.backend.dsl.models.requirement import Requirement
from strictdoc.backend.dsl.models.section import Section


class LinkRenderer:

    def __init__(self, output_html_root):
        self.output_html_root = output_html_root

    def render(self, node):
        if isinstance(node, Section):
            return self._string_to_link(node.title)
        if isinstance(node, Requirement):
            if node.title and len(node.title) > 0:
                return self._string_to_link(node.title)
            if node.uid and len(node.uid) > 0:
                return self._string_to_link(node.uid)

            # TODO: This is not reliable
            return str(id(node))
        raise NotImplementedError

    @staticmethod
    def _string_to_link(string):
        return re.sub(r'[^\w0-9]+', '-', string)



import re


def atoi(text):
    return int(text) if text.isdigit() else text


def alphanumeric_sort(text):
    '''
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    '''
    return [ atoi(c) for c in re.split(r'(\d+)', text) ]


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['strictdoc',
 'strictdoc.backend',
 'strictdoc.backend.dsl',
 'strictdoc.backend.dsl.models',
 'strictdoc.cli',
 'strictdoc.core',
 'strictdoc.core.actions',
 'strictdoc.export.html.generators',
 'strictdoc.export.html.renderers',
 'strictdoc.export.rst',
 'strictdoc.helpers']

package_data = \
{'': ['*'],
 'strictdoc': ['export/html/static/*',
               'export/html/templates/*',
               'export/html/templates/_shared/*',
               'export/html/templates/document_tree/*',
               'export/html/templates/single_document/*',
               'export/html/templates/single_document_table/*',
               'export/html/templates/single_document_traceability/*',
               'export/html/templates/single_document_traceability_deep/*']}

install_requires = \
['docutils==0.16', 'jinja2==2.11.2', 'textx==2.2.0']

entry_points = \
{'console_scripts': ['strictdoc = strictdoc.cli.main:main']}

setup_kwargs = {
    'name': 'strictdoc',
    'version': '0.0.3',
    'description': ' Software for writing technical requirements and specifications.',
    'long_description': '# StrictDoc\n\n![](https://github.com/stanislaw/strictdoc/workflows/StrictDoc%20on%20macOS/badge.svg)\n![](https://github.com/stanislaw/strictdoc/workflows/StrictDoc%20on%20Linux/badge.svg)\n![](https://github.com/stanislaw/strictdoc/workflows/StrictDoc%20on%20Windows/badge.svg)\n\nThe project and the repository are under construction.\n',
    'author': 'Stanislav Pankevich',
    'author_email': 's.pankevich@gmail.com',
    'maintainer': 'Stanislav Pankevich',
    'maintainer_email': 's.pankevich@gmail.com',
    'url': 'https://github.com/stanislaw/strictdoc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from strictdoc.core.document_meta import DocumentMeta


class Document(object):
    def __init__(self, name, free_texts, section_contents):
        self.name = name
        self.free_texts = free_texts
        self.section_contents = section_contents

        self.ng_sections = []
        self.ng_level = 0
        self.meta = None

    def __str__(self):
        return "Document: <name: {}, section_contents: {}>".format(
            self.name, self.section_contents
        )

    def __repr__(self):
        return self.__str__()

    def assign_meta(self, meta):
        assert isinstance(meta, DocumentMeta)
        self.meta = meta
